from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Union

from rich import print

if TYPE_CHECKING:
    from types import TracebackType

    from ..stack.manipulator import QueueManipulator
    from ..stack.navigator import QueueNavigator


@dataclass
class QueueOperation:
    stack_manipulator: "QueueManipulator"
    stack_navigator: "QueueNavigator"
    sync_time: Literal["enter", "exit", "none"] = "enter"
    sync_pull_requests: bool = True

    def __call__(
        self, sync_pull_requests: bool, sync_time: Literal["enter", "exit", "none"] = "enter"
    ) -> "QueueOperation":
        self.sync_pull_requests = sync_pull_requests
        self.sync_time = sync_time
        return self

    def __enter__(self):
        print(":arrows_clockwise: [bold green]Syncing with remote...[/bold green]")
        if self.sync_time == "enter":
            self.stack_manipulator.sync(sync_pull_requests=self.sync_pull_requests)

    def __exit__(
        self, exc_type: Union[type, None], exc_val: Exception, exc_tb: "TracebackType"
    ) -> None:
        if self.sync_time == "exit" and exc_type is not None:
            self.stack_manipulator.sync(sync_pull_requests=self.sync_pull_requests)
        self.stack_navigator.log()



from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

from lumberman.issues.provider import RemoteIssue

if TYPE_CHECKING:
    from .provider import Issue


@dataclass(frozen=True)
class IssueInfo:
    branch_title: str
    first_commit_str: str


class IssueStringifyer(Protocol):
    def get_issue_info(self, issue: "Issue") -> IssueInfo:
        ...


def sanitise_text_for_bash(input_string: str) -> str:
    char_to_remove = ["`"]
    for character in char_to_remove:
        input_string = input_string.replace(character, "")
    return input_string


class DefaultIssueStringifyer(IssueStringifyer):
    def get_issue_info(self, issue: "Issue") -> IssueInfo:
        return IssueInfo(
            branch_title=self._get_branch_title(issue=issue),
            first_commit_str=self._get_first_commit_str(issue=issue),
        )

    def _get_branch_title(self, issue: "Issue") -> str:
        branch_title: str = ""
        branch_title += f"{issue.title.prefix}/" if issue.title.prefix is not None else ""
        branch_title += f"{issue.entity_id}/" if isinstance(issue, RemoteIssue) else ""
        branch_title += issue.title.content

        return sanitise_text_for_bash(branch_title)

    def _get_first_commit_str(self, issue: "Issue") -> str:
        first_commit_str = issue.title.prefix if issue.title.prefix is not None else ""

        if isinstance(issue, RemoteIssue):
            first_commit_str += f"(#{issue.entity_id})"

        first_commit_str += ": " if first_commit_str else first_commit_str
        first_commit_str += issue.title.content

        if isinstance(issue, RemoteIssue):
            first_commit_str += f"""

Fixes #{issue.entity_id}"""

        return sanitise_text_for_bash(first_commit_str)



import pytest

from .provider import GithubIssueProvider


@pytest.mark.skip()
def test_issue_model():
    service = GithubIssueProvider()
    service.get_issues_assigned_to_me(in_progress_label="in progress")



from dataclasses import dataclass

import pytest

from .title_parser import IssueTitle, parse_issue_title


@dataclass(frozen=True)
class ParseIssueTitleExample:
    input_value: str
    expected: IssueTitle


@pytest.mark.parametrize(
    ("example"),
    [
        ParseIssueTitleExample(
            input_value="test(#77): test description",
            expected=IssueTitle(prefix="test", content="test description"),
        ),
        ParseIssueTitleExample(
            input_value="test: test description",
            expected=IssueTitle(prefix="test", content="test description"),
        ),
        ParseIssueTitleExample(
            input_value="no prefix", expected=IssueTitle(prefix=None, content="no prefix")
        ),
    ],
)
def test_issue_title_parsing(example: ParseIssueTitleExample):
    assert example.expected == parse_issue_title(example.input_value)



from dataclasses import dataclass

import pytest

from .provider import GithubIssue, Issue, LocalIssue
from .stringifyer import DefaultIssueStringifyer
from .title_parser import IssueTitle


@dataclass(frozen=True)
class TestIssueStringifyer:
    input_issue: Issue
    first_commit_str: str
    branch_title: str


@pytest.mark.parametrize(
    ("example"),
    [
        TestIssueStringifyer(
            input_issue=GithubIssue(
                entity_id="42",
                title=IssueTitle(prefix="test-prefix", content="test-description"),
                description="test-description",
            ),
            first_commit_str="""test-prefix(#42): test-description

Fixes #42""",
            branch_title="test-prefix/42/test-description",
        ),
        TestIssueStringifyer(
            input_issue=LocalIssue(title=IssueTitle(prefix=None, content="test-description")),
            first_commit_str="""test-description""",
            branch_title="test-description",
        ),
        TestIssueStringifyer(
            input_issue=GithubIssue(
                entity_id="42",
                title=IssueTitle(prefix="feat", content="test-description"),
                description="test-description",
            ),
            first_commit_str="""feat(#42): test-description

Fixes #42""",
            branch_title="feat/42/test-description",
        ),
        TestIssueStringifyer(
            input_issue=LocalIssue(
                title=IssueTitle(prefix="test-prefix", content="test-description")
            ),
            first_commit_str="""test-prefix: test-description""",
            branch_title="test-prefix/test-description",
        ),
    ],
)
def test_default_issue_stringifyer(example: TestIssueStringifyer):
    issue_info = DefaultIssueStringifyer().get_issue_info(issue=example.input_issue)
    assert issue_info.first_commit_str == example.first_commit_str
    assert issue_info.branch_title == example.branch_title



from lumberman.cli.config import STACK_NAVIGATOR
from lumberman.cli.location import FullLocation, LocationCLIOption

from ..git import StagingMigrater
from .config import STACK_OP


def trunk():
    """Go to the trunk of the stack."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_NAVIGATOR.trunk()


def bottom():
    """Go to the bottom of the stack."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_NAVIGATOR.bottom()


def down():
    """Go to the item below the current one."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_NAVIGATOR.down()


def up():
    """Go to the item above the current one."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_NAVIGATOR.up()


def top():
    """Go to the top of the stack."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_NAVIGATOR.top()


def log():
    """Print the current stack status."""
    STACK_NAVIGATOR.log()


def checkout():
    """Prompt to checkout an item in the stack."""
    with StagingMigrater():
        STACK_NAVIGATOR.checkout()


def navigate_to_insert_location(location: LocationCLIOption):
    match location.to_full_location:
        case FullLocation.trunk:
            STACK_NAVIGATOR.trunk()
        case FullLocation.bottom:
            STACK_NAVIGATOR.bottom()
        case FullLocation.top:
            STACK_NAVIGATOR.top()
        case FullLocation.up:
            pass
        case FullLocation.down:
            STACK_NAVIGATOR.down()



from dataclasses import dataclass
from typing import Protocol

from rich import box, print
from rich.panel import Panel

from ..cli.subprocess_utils import interactive_cmd, shell_output
from ..git import StagingMigrater


class QueueNavigator(Protocol):
    def trunk(self):
        ...

    def bottom(self):
        ...

    def top(self):
        ...

    def down(self):
        ...

    def up(self):
        ...

    def log(self):
        ...

    def checkout(self):
        ...


@dataclass(frozen=True)
class GraphiteNavigator(QueueNavigator):
    def trunk(self):
        with StagingMigrater():
            interactive_cmd("gt trunk")

    def bottom(self):
        with StagingMigrater():
            interactive_cmd("gt bottom")

    def top(self):
        with StagingMigrater():
            interactive_cmd("gt top")

    def down(self):
        interactive_cmd("gt down")

    def up(self):
        interactive_cmd("gt up")

    def log(self):
        result: str = shell_output("gt log short")  # type: ignore
        print(
            "\n",
            Panel(result, title="Top", subtitle="Bottom", expand=False, box=box.HORIZONTALS),
            "\n",
        )

    def checkout(self):
        interactive_cmd("gt checkout")



from subprocess import STDOUT
from typing import Optional


def interactive_cmd(command: str) -> None:
    import subprocess

    try:
        subprocess.run(command, shell=True, stderr=STDOUT, check=True)
    except subprocess.CalledProcessError as e:
        if not e.stdout:
            return

        error_message = f"""{command} failed.
\tExit code: {e.returncode}"""

        if e.stdout:
            error_message += f"\n\tOutput: {e.stdout.decode('utf-8').strip()}"

        raise RuntimeError(error_message) from e


def shell_output(command: str) -> Optional[str]:
    import subprocess

    try:
        result = subprocess.check_output(command, shell=True, stderr=STDOUT).decode("utf-8").strip()
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"""{command} failed.
    Exit code: {e.returncode}
    Output: {e.stdout.decode('utf-8').strip()}"""
        ) from e

    if not result:
        return None

    return result



import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Mapping, Optional, Protocol, runtime_checkable

from lumberman.cli.subprocess_utils import shell_output
from lumberman.issues.title_parser import IssueTitle, parse_issue_title

if TYPE_CHECKING:
    from collections.abc import Sequence


@dataclass
class IssueComment:
    id: str
    author_login: str
    body: str
    url: str


def _parse_issue_comment(comment_json: Mapping[str, str]) -> IssueComment:
    return IssueComment(
        id=comment_json["id"],  # type: ignore
        body=comment_json["body"],  # type: ignore
        url=comment_json["url"],  # type: ignore
        author_login=comment_json["author"]["login"],  # type: ignore
    )


def _create_label(label: str) -> None:
    shell_output(f"gh label create {label}")


@runtime_checkable
class Issue(Protocol):
    title: IssueTitle


@dataclass(frozen=True)
class LocalIssue(Issue):
    """Issue created by entering a title in the CLI"""

    title: IssueTitle


@runtime_checkable
class RemoteIssue(Protocol):
    """Issue created by a remote source (e.g. Github)"""

    entity_id: str
    description: str

    def label(self, label: str) -> None:
        ...

    def assign(self, assignee: str) -> None:
        ...

    def get_comments(self) -> "Sequence[IssueComment]":
        ...


@dataclass(frozen=True)
class GithubIssue(Issue, RemoteIssue):
    entity_id: str
    title: IssueTitle
    description: str

    def _add_label(self, label: str) -> None:
        if not self.entity_id:
            return

        shell_output(f"gh issue edit {int(self.entity_id)} --add-label {label}")

    def label(self, label: str) -> None:
        if not self.entity_id:
            return

        try:
            self._add_label(label)
        except Exception:
            try:
                _create_label(label)
                self._add_label(label)
            except Exception as e:
                raise RuntimeError(f"Error labeling issue {self.entity_id} with {label}") from e

    def get_comments(self) -> "Sequence[IssueComment]":
        if not self.entity_id:
            return []

        comments_json = shell_output(f"gh issue view {self.entity_id} --json comments")
        comments: Sequence[Mapping[str, str]] = json.loads(comments_json)["comments"]  # type: ignore
        return [_parse_issue_comment(c) for c in comments]

    def assign(self, assignee: str) -> None:
        if not self.entity_id:
            return

        shell_output(f"gh issue edit {int(self.entity_id)} --add-assignee {assignee}")


class IssueProvider(Protocol):
    def setup(self) -> None:
        """Any setup needed, including installing CLI tools, etc."""
        ...

    def get_latest_issues(self, in_progress_label: str) -> "Sequence[GithubIssue]":
        ...

    def get_issues_assigned_to_me(self, in_progress_label: str) -> "Sequence[GithubIssue]":
        ...

    def get_current_issue(self) -> Optional[GithubIssue]:
        ...


class GithubIssueProvider(IssueProvider):
    def setup(self) -> None:
        pass

    def _values_to_issue(self, values: dict[str, str]) -> GithubIssue:
        parsed_title = parse_issue_title(values["title"])
        return GithubIssue(
            entity_id=str(values["number"]), title=parsed_title, description=values["body"]
        )

    def get_latest_issues(self, in_progress_label: str) -> "Sequence[GithubIssue]":
        latest_issues = shell_output(
            f"gh issue list --limit 10 --json number,title,body --search 'is:open -label:{in_progress_label}'"
        )

        if latest_issues is None:
            return []

        return self._parse_github_json_str(latest_issues)

    def get_issues_assigned_to_me(self, in_progress_label: str) -> "Sequence[GithubIssue]":
        """Get issues assigned to current user on current repo"""
        my_issues_cmd = shell_output(
            f"gh issue list --assignee @me  --search '-label:{in_progress_label}' --json number,title,body"
        )

        if my_issues_cmd is None:
            return []

        return self._parse_github_json_str(my_issues_cmd)

    def _parse_github_json_str(self, issue_str: str) -> "Sequence[GithubIssue]":
        values = json.loads(issue_str)
        parsed_output = [self._values_to_issue(v) for v in values]
        return parsed_output

    def get_current_issue(self) -> Optional[GithubIssue]:
        current_branch: str = shell_output("git rev-parse --abbrev-ref HEAD")  # type: ignore

        branch_items = current_branch.split("/")

        if len(branch_items) != 3:
            return None

        return GithubIssue(
            entity_id=branch_items[1],
            title=IssueTitle(prefix=branch_items[0], content=branch_items[2]),
            description="",
        )


if __name__ == "__main__":
    GithubIssue(
        entity_id="257",
        title=IssueTitle(prefix="test-prefix", content="test-description"),
        description="test-description",
    ).get_comments()



import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

import typer

from lumberman.cli import manipulation as man
from lumberman.cli import navigation as nav

if TYPE_CHECKING:
    from collections.abc import Sequence

app = typer.Typer(
    name="[l]umber[m]an",
    no_args_is_help=True,
    add_help_option=False,
    add_completion=False,
    help="All commands are registered as [sh]orthand. You can call the command as 'lm sh' or 'lumberman shorthand'.",
)


@dataclass(frozen=True)
class Command:
    name: str
    fn: Callable[[], None]


@dataclass(frozen=True)
class CommandSection:
    name: str
    commands: "Sequence[Command]"


commands = [
    CommandSection(
        name="Navigation",
        commands=[
            Command(name="[ch]heckout", fn=nav.checkout),
            Command(name="[bo]ttom", fn=nav.bottom),
            Command(name="[to]p", fn=nav.top),
            Command(name="[do]wn", fn=nav.down),
            Command(name="[up]", fn=nav.up),
        ],
    ),
    CommandSection(name="Orientation", commands=[Command(name="[l]og", fn=nav.log)]),
    CommandSection(
        name="Manipulation",
        commands=[
            Command(name="[i]nsert", fn=man.insert),
            Command(name="[j]ab", fn=man.jab),
            Command(name="[d]elete", fn=man.delete),
            Command(name="[f]ork", fn=man.fork),
            Command(name="[m]ove", fn=man.move),
            Command(name="[n]ew", fn=man.new),
            Command(name="[s]ync", fn=man.sync),
        ],
    ),
]

shorthands: set[str] = set()

# Add the commands and shorthands to the app
for section in commands:
    for command in section.commands:
        # Add the combined command, e.g. [i]nsert
        app.command(name=command.name, rich_help_panel=section.name)(command.fn)

        # Handle shorthand, e.g. i
        command_shorthand = re.findall(r"\[(.*?)\]", command.name)[0]
        if command_shorthand in shorthands:
            raise ValueError(f"Duplicate shorthand '{command_shorthand}'")

        shorthands.add(command_shorthand)
        app.command(name=command_shorthand, hidden=True)(command.fn)

        # Add the full command, e.g. insert
        command_full = command.name.replace("[", "").replace("]", "")
        app.command(name=command_full, hidden=True)(command.fn)


if __name__ == "__main__":
    app()



import re
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class IssueTitle:
    prefix: Optional[str]
    content: str

    def __str__(self) -> str:
        if self.prefix:
            return f"{self.prefix}: {self.content}"
        return f"{self.content}"


def sanitise_text_for_git(input_string: str) -> str:
    char2replacement = {
        " ": "-",
        ":": "/",
        ",": "",
        "'": "",
        '"': "",
        "(": "",
        ")": "",
        "[": "",
        "": "",
        "`": "",
        ">": "",
        "<": "",
        "=": "",
    }

    for character, replacement in char2replacement.items():
        input_string = input_string.replace(character, replacement)

    return input_string.replace("--", "-")


def parse_issue_title(issue_title: str) -> IssueTitle:
    # Get all string between start and first ":"
    try:
        prefix = re.findall(r"^(.*?)[\(:]", issue_title)[0]
    except IndexError:
        # No prefix found, return without prefix
        return IssueTitle(prefix=None, content=issue_title)

    description = re.findall(r": (.*)$", issue_title)[0]
    return IssueTitle(prefix=prefix, content=description)



from typing import TYPE_CHECKING

from rich import print

from lumberman.cli.config import ISSUE_CONTROLLER, STACK_MANIPULATOR, STACK_NAVIGATOR, STACK_OP
from lumberman.cli.location import FullLocation, Location, LocationCLIOption
from lumberman.cli.markdown import print_md
from lumberman.cli.navigation import navigate_to_insert_location
from lumberman.issues.provider import GithubIssue, Issue, RemoteIssue
from lumberman.issues.selecter import DefaultIssueSelecter

if TYPE_CHECKING:
    from lumberman.issues.provider import IssueComment


def _markdown_quote_string(string: str) -> str:
    return "\n".join([f"> {line}" for line in string.split("\n")])


def _print_comment(comment: "IssueComment"):
    print_md(f"## {comment.author_login}")
    print_md(_markdown_quote_string(comment.body))


def _select_issue() -> "Issue":
    selected_issue = ISSUE_CONTROLLER.select_issue()
    print_md(f"# {selected_issue.title!s}")

    if isinstance(selected_issue, GithubIssue):
        print_md("## Description")
        description = _markdown_quote_string(selected_issue.description)
        print_md(description)

        if comments := selected_issue.get_comments():
            for comment in comments:
                _print_comment(comment)

    return selected_issue


def jab(location: LocationCLIOption = Location.up):
    """Prompt to create a new item on the current stack, without fetching issues."""
    with STACK_OP(sync_time="exit", sync_pull_requests=False):
        selected_issue = DefaultIssueSelecter().select_issue_dialog([])
        navigate_to_insert_location(location)
        STACK_MANIPULATOR.insert(selected_issue)


def insert(location: LocationCLIOption = Location.up):
    """Prompt to create a new item on the current stack. Defaults to creating an item in between the current item and the next item."""
    with STACK_OP(sync_time="exit", sync_pull_requests=False):
        selected_issue = _select_issue()

        navigate_to_insert_location(location)

        STACK_MANIPULATOR.insert(selected_issue)
        if isinstance(selected_issue, RemoteIssue):
            selected_issue.label("in-progress")
            selected_issue.assign(assignee="@me")


def move():
    """Move the current item to a new location in the stack."""
    with STACK_OP(sync_time="exit", sync_pull_requests=True):
        STACK_MANIPULATOR.move()


def delete():
    """Prompt to delete an item."""
    with STACK_OP(sync_time="exit", sync_pull_requests=False):
        STACK_MANIPULATOR.delete()


def fork(location: LocationCLIOption = Location.bottom):
    """Fork into a new stack and add an item. Defaults to forking from the first item in the current stack."""
    with STACK_OP(sync_time="enter", sync_pull_requests=False):
        selected_issue = _select_issue()

        match location.to_full_location:
            case FullLocation.bottom | FullLocation.trunk:
                STACK_NAVIGATOR.bottom()
                STACK_NAVIGATOR.up()
            case FullLocation.top:
                STACK_NAVIGATOR.top()
                STACK_NAVIGATOR.down()
            case FullLocation.up:
                pass  # No need to do anything, already in the correct location
            case FullLocation.down:
                STACK_NAVIGATOR.down()

        STACK_MANIPULATOR.fork(selected_issue)
        if isinstance(selected_issue, RemoteIssue):
            selected_issue.label("in-progress")
            selected_issue.assign(assignee="@me")


def new():
    """Start a new stack on top of trunk."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        selected_issue = _select_issue()
        STACK_MANIPULATOR.sync(sync_pull_requests=False)
        STACK_NAVIGATOR.trunk()
        STACK_MANIPULATOR.fork(selected_issue)

        if isinstance(selected_issue, RemoteIssue):
            selected_issue.label("in-progress")
            selected_issue.assign(assignee="@me")


def sync(
    automerge: bool = False, draft: bool = False, squash: bool = False, add_pr_label: bool = True
):
    """Synchronize all state, ensuring the stack is internally in sync, and in sync with the remote. Creates PRs if needed."""
    with STACK_OP(sync_time="none", sync_pull_requests=False):
        STACK_MANIPULATOR.sync(
            automerge=automerge, squash=squash, draft=draft, sync_pull_requests=True
        )

        if add_pr_label:
            current_issue = ISSUE_CONTROLLER.provider.get_current_issue()
            if current_issue:
                current_issue.label("has-pr")
        print(":rocket: [bold green]Stack synced![/bold green]")


if __name__ == "__main__":
    insert()



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from rich.progress import Progress, SpinnerColumn, TextColumn

if TYPE_CHECKING:
    from collections.abc import Sequence

    from lumberman.issues.provider import Issue

    from .provider import GithubIssue, IssueProvider
    from .selecter import IssueSelecter


@dataclass(frozen=True)
class IssueController:
    view: "IssueSelecter"
    provider: "IssueProvider"
    in_progress_label: str

    def _get_latest_issues(self) -> "Sequence[GithubIssue]":
        with Progress(
            SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
        ) as progress:
            progress.add_task("Getting latest issues", start=True)
            latest_issues = self.provider.get_latest_issues(
                in_progress_label=self.in_progress_label
            )

        if not latest_issues:
            return []

        return latest_issues

    def _get_my_issues(self) -> "Sequence[GithubIssue]":
        with Progress(
            SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
        ) as progress:
            progress.add_task("Getting issues assigned to you", start=True)
            my_issues = self.provider.get_issues_assigned_to_me(
                in_progress_label=self.in_progress_label
            )

        return my_issues

    def select_issue(self, issues: Optional["Sequence[GithubIssue]"] = None) -> "Issue":
        if not issues:
            issues = [*self._get_my_issues(), *self._get_latest_issues()]

        selected_issue = self.view.select_issue_dialog(issues)

        return selected_issue



from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

from lumberman.issues.provider import Issue

from ..cli.subprocess_utils import interactive_cmd

if TYPE_CHECKING:
    from ..issues.provider import Issue
    from ..issues.stringifyer import IssueStringifyer


class QueueManipulator(Protocol):
    issue_parser: "IssueStringifyer"

    def fork(self, issue: "Issue"):
        """Create a new item, forking from the current item."""
        ...

    def insert(self, issue: "Issue"):
        """Create a new item. If an item exists above the current item, insert between them."""
        ...

    def delete(self):
        ...

    def move(self):
        ...

    def sync(
        self,
        sync_pull_requests: bool = True,
        automerge: bool = False,
        squash: bool = False,
        draft: bool = True,
    ):
        ...


@dataclass(frozen=True)
class GraphiteManipulator(QueueManipulator):
    issue_parser: "IssueStringifyer"

    def _new_branch(self, insert: bool, issue: "Issue"):
        issue_info = self.issue_parser.get_issue_info(issue)

        create_command = f'gt create "{issue_info.branch_title}" -m "{issue_info.first_commit_str}" --no-interactive'
        create_command += " --insert" if insert else ""

        print(f"Creating branch with command: {create_command}")
        interactive_cmd(create_command)
        interactive_cmd("git add -A")
        interactive_cmd(f'git commit --allow-empty -m "{issue_info.first_commit_str}"')

    def fork(self, issue: "Issue"):
        self._new_branch(insert=False, issue=issue)

    def insert(self, issue: "Issue"):
        self._new_branch(insert=True, issue=issue)

    def delete(self):
        interactive_cmd("gt delete")

    def move(self):
        interactive_cmd("gt move")

    def sync(
        self,
        sync_pull_requests: bool = True,
        automerge: bool = False,
        squash: bool = False,
        draft: bool = True,
    ):
        if squash:
            interactive_cmd("gt squash --no-edit")

        interactive_cmd("gt sync --pull --force --restack --delete")

        if sync_pull_requests:
            command = "gt submit --no-edit --stack"
            if draft:
                command += " --draft"
            else:
                command += " --publish"

            if automerge:
                command += " --merge-when-ready"
            interactive_cmd(command)



import enum
from typing import Annotated

import typer


class FullLocation(str, enum.Enum):
    trunk = "trunk"
    bottom = "bottom"
    top = "top"
    up = "up"
    down = "down"


class Location(str, enum.Enum):
    up = "up"
    top = "top"
    to = "to"
    down = "down"
    do = "do"
    bottom = "bottom"
    bo = "bo"
    trunk = "trunk"
    tr = "tr"

    @property
    def to_full_location(self) -> "FullLocation":
        match self:
            case Location.to | Location.top:
                return FullLocation.top
            case Location.do | Location.down:
                return FullLocation.down
            case Location.bo | Location.bottom:
                return FullLocation.bottom
            case Location.tr | Location.trunk:
                return FullLocation.trunk
            case Location.up:
                return FullLocation.up


LocationCLIOption = Annotated[Location, typer.Argument()]



# ############################
# ## NOTES ON IMPORT FORMAT ##
# ############################
#
# From https://github.com/dagster-io/dagster/blob/master/python_modules/dagster/dagster/__init__.py
#
# This file defines your package's public API. Imports need to be structured/formatted so as to to ensure
# that the broadest possible set of static analyzers understand your_package's public API as intended.
# The below guidelines ensure this is the case.
#
# (1) All imports in this module intended to define exported symbols should be of the form `from
# your_package.foo import X as X`. This is because imported symbols are not by default considered public
# by static analyzers. The redundant alias form `import X as X` overwrites the private imported `X`
# with a public `X` bound to the same value. It is also possible to expose `X` as public by listing
# it inside `__all__`, but the redundant alias form is preferred here due to easier maintainability.

# (2) All imports should target the module in which a symbol is actually defined, rather than a
# container module where it is imported. This rule also derives from the default private status of
# imported symbols. So long as there is a private import somewhere in the import chain leading from
# an import to its definition, some linters will be triggered (e.g. pyright). For example, the
# following results in a linter error when using your_package as a third-party library:

#     ### your_package/foo/bar.py
#     BAR = "BAR"
#
#     ### your_package/foo/__init__.py
#     from .bar import BAR  # BAR is imported so it is not part of your_package.foo public interface
#     FOO = "FOO"
#
#     ### your_package/__init__.py
#     from .foo import FOO, BAR  # importing BAR is importing a private symbol from your_package.foo
#     __all__ = ["FOO", "BAR"]
#
#     ### some_user_code.py
#     # from your_package import BAR  # linter error even though `BAR` is in `your_package.__all__`!
#
# We could get around this by always remembering to use the `from .foo import X as X` form in
# containers, but it is simpler to just import directly from the defining module.



from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

import typer
from iterfzf import iterfzf  # type: ignore
from rich.console import Console

from .provider import GithubIssue, Issue, LocalIssue
from .title_parser import parse_issue_title

if TYPE_CHECKING:
    from collections.abc import Sequence

console = Console()


class IssueSelecter(Protocol):
    def select_issue_dialog(self, issues: "Sequence[GithubIssue]") -> Issue:
        ...


@dataclass(frozen=True)
class FZFSelection:
    input_str: str  # String typed into FZF dialog
    selected_str: str | None  # Selected string from FZF, e.g. the matching entry

    def either(self) -> str:
        return self.selected_str or self.input_str


@dataclass(frozen=True)
class DefaultIssueSelecter(IssueSelecter):
    def _show_selection_dialog(self, issues: "Sequence[GithubIssue]") -> FZFSelection:
        issue_titles = [f"{issue.title.content} #{issue.entity_id}" for issue in issues]
        if len(issue_titles) == 0:
            issue_titles.append("No issues found. Enter an title for the change below.")

        typer.echo("Select an issue or enter a new issue title.")
        selection: tuple(str, str) = iterfzf([*issue_titles], print_query=True, exact=True)  # type: ignore
        return FZFSelection(input_str=selection[0], selected_str=selection[1])  # type: ignore

    def select_issue_dialog(self, issues: "Sequence[GithubIssue]") -> Issue:
        fzf_selection = self._show_selection_dialog(issues=issues)
        selected_issue_title = fzf_selection.either()

        selected_issue_from_list = [i for i in issues if i.title.content in selected_issue_title]
        if selected_issue_from_list:
            return next(issue for issue in selected_issue_from_list)

        parsed_title = parse_issue_title(selected_issue_title)
        return LocalIssue(title=parsed_title)



from ..issues.controller import IssueController
from ..issues.provider import GithubIssueProvider
from ..issues.selecter import DefaultIssueSelecter
from ..issues.stringifyer import DefaultIssueStringifyer
from ..stack.manipulator import GraphiteManipulator
from ..stack.navigator import GraphiteNavigator
from .interface_elements import QueueOperation

STACK_MANIPULATOR = GraphiteManipulator(issue_parser=DefaultIssueStringifyer())
STACK_NAVIGATOR = GraphiteNavigator()
ISSUE_CONTROLLER = IssueController(
    provider=GithubIssueProvider(), view=DefaultIssueSelecter(), in_progress_label="in-progress"
)
STACK_OP = QueueOperation(
    sync_time="enter", stack_manipulator=STACK_MANIPULATOR, stack_navigator=STACK_NAVIGATOR
)
